/*
    tcl-mmap - Tcl Interface to POSIX mmap

    Copyright (C) 2008  Alexandros Stergiakis

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * mmap.c
 *
 */

#include <sys/mman.h> // mmap, msync, munmap
#include <stdio.h>  // SEEK_ constants 
#include <sys/types.h> // fstat
#include <sys/stat.h> // fstat
#include <unistd.h> // fstat, sysconf
#include <string.h> // memcpy
#include <signal.h> // signal

#include <tcl.h>

/*
 * Function Prototypes
 */

static int MmapCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

void NewChannelName(char *name, CONST char* prefix);

static int Input(ClientData instanceData, char *buf, int bufSize, int *errorCodePtr);

static int Output(ClientData instanceData, const char *buf, int bufSize, int *errorCodePtr);

static int Seek(ClientData instanceData, long offset, int seekMode, int *errorCodePtr);

static int Close(ClientData instanceData, Tcl_Interp *interp);

static void Watch(ClientData instanceData, int mask);

//static int BlockMode(ClientData instanceData, int mode);

static int GetHandle(ClientData instanceData, int direction, ClientData *handlePtr);

//static void catch_signal(int sig_num);

/*
 * This structure describes the channel type structure for in-memory channels:
 */

static Tcl_ChannelType channelType = {
        "mmap", /* Channel type name */
        TCL_CHANNEL_VERSION_2,
        (Tcl_DriverCloseProc *) Close,
        (Tcl_DriverInputProc *) Input,
        (Tcl_DriverOutputProc *) Output,
        (Tcl_DriverSeekProc *) Seek,
        (Tcl_DriverSetOptionProc *) NULL, /* no channel type specific options */
        (Tcl_DriverGetOptionProc *) NULL, /* no channel type specific options */
        (Tcl_DriverWatchProc *) Watch,
        (Tcl_DriverGetHandleProc *) GetHandle, 
        (Tcl_DriverClose2Proc *) NULL, /* no support for closing the read and write sides independently */
        (Tcl_DriverBlockModeProc *) NULL, /* mmap channels are always in non-blocking mode */
        //(Tcl_DriverBlockModeProc *) BlockMode,
        (Tcl_DriverFlushProc *) NULL, /* currently reserved for future use. It should be set to NULL */
        (Tcl_DriverHandlerProc *) NULL  /* this is not a stacked channel driver */
}; 

/*
 * This structure describes the per-instance state of a channel.
 */

typedef struct ChannelInstance {
  char*          addr;      /* Address where the mapped region starts */
  unsigned long  seek;      /* Current location to read from (or write to). */
  size_t         length;    /* Number of allocated bytes */
  int            prot;      /* Protection of the mapped region: PROT_READ, PROT_WRITE, PROT_EXEC, PROT_NONE */
  int            flags;     /* Attributes of the mapped region: MAP_FIXED, MAP_SHARED, MAP_PRIVATE */
  int            fd;        /* The file descriptor of the mapped file */
  off_t          offset;    /* The starting offset in the file to map from */
  Tcl_Channel    chan;      /* Backreference to generic channel information */
} ChannelInstance;

//Tcl_Interp *mainInterp;

/*
 * Function Bodies
 */
/* Notes:
 *   - Reaching the end of mapped memory region, generates an EOF.
 *   - An mmapped region is automatically unmapped when the process terminates.
 *   - Seek value wraps around to point to the start of the mmapped region, when 
 *     it reaches the end of it.
 *   - The value to -offset is required to multiples of the system's virtual
 *     memory page size. This can be retrieved (from C) with long sysconf(_SC_PAGESIZE)
 *   - Writes on memory mapped area which are outside of the actual file size,
 *     has no effect on the file itself. This can happen due to the offset
 *     requirement above.
 *   - A memory-mapped region is inherited by a child across a fork, but not
 *     across an exec.
 *   - Writes are not reflected to the actual file at once. They are handled 
 *     automatically by the kernel's virtual memory algorithm. Other processes 
 *     however will always get the latest content.
 *   - When -anonymous is specified, memory mapped area is initialized to zero.
 *     (not implemented)
 *   - If two (or more) related processes map the same area of /dev/null, then this
 *     area (or the overlappting part) is shared memory among the processes, and
 *     it is initialized to zero.
 */

static int Input(ClientData instanceData, char *buf, int bufSize, int *errorCodePtr) {
    ChannelInstance* data;
    data = (ChannelInstance*) instanceData;

    if (data->seek >= data->length) {
        return 0; // EOF
    }

    if (bufSize > data->length - data->seek) {
        bufSize = data->length - data->seek;
    }

    memcpy((char *)buf, (char *) (data->addr + data->seek), (size_t) bufSize);
    data->seek += bufSize;

    return bufSize;
}

static int Output(ClientData instanceData, const char *buf, int bufSize, int *errorCodePtr) {
    ChannelInstance* data;
    data = (ChannelInstance*) instanceData;
    int ret;

    if (bufSize > data->length - data->seek) {
        bufSize = data->length - data->seek;
        memcpy((char *) (data->addr + data->seek), (char *) buf, (size_t) bufSize);
        data->seek = 0;
    } else {
        memcpy((char *) (data->addr + data->seek), (char *) buf, (size_t) bufSize);
        data->seek += bufSize;
    }
    ret = msync(data->addr, data->length, MS_SYNC); // also possible: MS_ASYNC & MS_INVALIDATE 
    if (ret == -1) {
        *errorCodePtr = Tcl_GetErrno();
        return -1;
    }

    return bufSize;
}

static int Seek(ClientData instanceData, long offset, int seekMode, int *errorCodePtr) {
    ChannelInstance* data;
    data = (ChannelInstance*) instanceData;

/*
 From seek(n):
       start     The new access position will be offset bytes from the start of the underlying file or device.

       current   The  new access position will be offset bytes from the current access position; a negative offset moves the access position backwards in
                 the underlying file or device.

       end       The new access position will be offset bytes from the end of the file or device.  A negative offset places the  access  position  before
                 the end of file, and a positive offset places the access position after the end of file.
*/

    switch (seekMode) {
    case SEEK_SET:
        data->seek = offset;
        break;

    case SEEK_CUR:
        data->seek += offset;
        break;

    case SEEK_END:
        data->seek = data->length + offset;
        break;
    }

    /* We check if seek is within range, and fix */
    if (data->seek < 0) data->seek = 0;
    if (data->seek > data->length) data->seek = data->length;
    
    return data->seek;
}

static int Close(ClientData instanceData, Tcl_Interp *interp) {
    int result = 0; /* for success */
    int ret;

    ChannelInstance* data;
    data = (ChannelInstance*) instanceData;

    ret = msync(data->addr, data->length, MS_SYNC); // also possible: MS_ASYNC & MS_INVALIDATE 
    if (ret == -1) {
        Tcl_AppendResult (interp, "msync error: ", Tcl_PosixError(interp), "; ", NULL);
        result = -1;
    }
    ret = munmap(data->addr, data->length);
    if (ret == -1) {
        Tcl_AppendResult (interp, "munmap error: ", Tcl_PosixError(interp), NULL);
        result = -1;
    }
    ckfree((char *) data);

    return result;
}

static void Watch(ClientData instanceData, int mask) {
    /* we never call Tcl_NotifyChannel since mmapped *
     * files are always readable and writable        */
}

/*
static int BlockMode(ClientData instanceData, int mode) {
    // mmap channels are always in non-blocking mode 
    return 0;
}
*/

static int GetHandle(ClientData instanceData, int direction, ClientData *handlePtr) {
    /* mmap channels have no associated file handle descriptor */
    return TCL_ERROR;
}

/*
static void catch_signal(int sig_num) {
    signal(sig_num, catch_signal);
    switch (sig_num) {
        case SIGSEGV:
            Tcl_SetResult(mainInterp, "Attempted write into a region mapped as read-only", TCL_STATIC);
            break;
        case SIGBUS:
            Tcl_SetResult(mainInterp, "Attempted  access  to  a portion of the buffer that does not correspond to the file", TCL_STATIC);
            break;
    }
    Tcl_BackgroundError(mainInterp);
}
*/

/*
 * mmap ?-shared? ?-length bytes? ?-offset bytes? channelId 
 */
static int MmapCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) {
    static char* cmds[] = { "-shared", "-length", "-offset", NULL };
    int index;
    int ret;

    /* Default values */
    int length = 0;
    int offset = 0;
    int flag = MAP_PRIVATE;
    int prot = PROT_NONE;

    if (objc < 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "?-shared? ?-length bytes? ?-offset bytes? channelId");
        return TCL_ERROR;
    }

    int i;
    for (i = 1; i < objc - 1; i++) {
        if (Tcl_GetIndexFromObj(interp, objv[i], cmds, "option", 0, &index) != TCL_OK)
            return TCL_ERROR;

        switch (index) {
        case 0: /* -shared */
            flag &= ~MAP_PRIVATE;
            flag = MAP_SHARED;
            break;
        case 1: /* -length */
            if (Tcl_GetInt(interp, Tcl_GetString(objv[++i]), &length) == TCL_ERROR)
                return TCL_ERROR;
            break;
        case 2: /* -offset */
            if (Tcl_GetInt(interp, Tcl_GetString(objv[++i]), &offset) == TCL_ERROR)
                return TCL_ERROR;

            long page_size = sysconf(_SC_PAGESIZE);
            if (offset % page_size != 0) {
                char page[50];
                sprintf (page, "%ld", (long) page_size);
                Tcl_AppendResult(interp, "Wrong offset value, it must be a multiple of ", page, NULL);
                return TCL_ERROR;
            }
            break;
        default:
            Tcl_AppendResult(interp, "Unrecognized option", Tcl_GetString(objv[i]), NULL);
            return TCL_ERROR;
        } /* switch */
    } /* for */
        
    if (i != objc - 1) {
        Tcl_SetResult(interp, "channelId is missing", TCL_STATIC);
        return TCL_ERROR;
    }

    Tcl_Channel chan;
    chan = Tcl_GetChannel(interp, Tcl_GetString(objv[i]), NULL);
    if (chan == NULL) {
        Tcl_AppendResult(interp, "Could not find channel ", Tcl_GetString(objv[i]), NULL);
        return TCL_ERROR;
    }

    int mode;
    mode = Tcl_GetChannelMode(chan);
    if (mode & TCL_READABLE) {
        prot &= ~PROT_NONE;
        prot |= PROT_READ;
    }
    if (mode & TCL_WRITABLE) {
        prot &= ~PROT_NONE;
        prot |= PROT_WRITE;
    }
    /* Note we don't use PROT_EXEC */

    int fd = 0;
    ClientData cd;
    ret = Tcl_GetChannelHandle(chan, mode, (ClientData *) &cd);
    fd = (int) cd; // This should be ok because the pointer size in bytes is at least as may bytes as 'int'
    if (ret != TCL_OK) {
        Tcl_SetResult(interp, "Could not get file handle", TCL_STATIC);
        return TCL_ERROR;
    }

    if (length == 0) {
        struct stat statbuf;
        if (fstat(fd, &statbuf) < 0) {
            Tcl_AppendResult (interp, "fstat error: ", Tcl_PosixError(interp), NULL);
            return TCL_ERROR;
        }
        length = statbuf.st_size - offset;
    }

    char *addr;
    addr = (char *) mmap(NULL, length, prot, flag, fd, offset);
    if (addr == MAP_FAILED) {
        Tcl_AppendResult (interp, "mmap error: ", Tcl_PosixError(interp), NULL);
        return TCL_ERROR;
    }

    //signal(SIGSEGV, catch_signal);
    //signal(SIGBUS, catch_signal);

    ChannelInstance *data;
    data = (ChannelInstance *) ckalloc(sizeof (struct ChannelInstance));
    if (data == NULL) {
        Tcl_SetResult(interp, "Memory allocation failed", TCL_STATIC);
        return TCL_ERROR;
    }

    data->addr = addr;
    data->seek = 0;
    data->length = (size_t) length;
    data->prot = prot;
    data->flags = flag;
    data->fd = fd;
    data->offset = (off_t) offset;

    char channelName [50];
    NewChannelName(channelName, "mmap");

    Tcl_Channel mmap_chan;
    mmap_chan = Tcl_CreateChannel(&channelType, channelName, (ClientData) data, mode);
    data->chan = mmap_chan;

    Tcl_RegisterChannel(interp, mmap_chan);

    Tcl_SetChannelOption(interp, mmap_chan, "-buffering", "none");

    Tcl_SetResult(interp, (char *) Tcl_GetChannelName(mmap_chan), TCL_VOLATILE);
    return TCL_OK;
}

void NewChannelName(char *name, CONST char* prefix) {
    TCL_DECLARE_MUTEX (mmapCounterMutex)
        static unsigned long mmapCounter = 0;

    Tcl_MutexLock (&mmapCounterMutex);
    {
        snprintf(name, 50, "%s%lu", prefix, (unsigned long) mmapCounter);
        mmapCounter++;
    }
    Tcl_MutexUnlock (&mmapCounterMutex);
}

int Mmap_Init(Tcl_Interp *interp) {
    if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {
        return TCL_ERROR;
    }

    Tcl_CreateObjCommand(interp, "mmap", MmapCmd, (ClientData) NULL, NULL);

    //mainInterp = interp;

    Tcl_PkgProvide(interp, "mmap", "1.1");
    return TCL_OK;
}

